//	BGtemplate CodeWarrior file for developing max/msp externals
//		-- many thanks to Dan Trueman for the orginal!
//			Brad Garton, 11/2005


#include "ext.h"
#include "z_dsp.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define MAX_INPUTS 10 	// arbitrary
#define MAX_OUTPUTS 10	// also arbitrary

void *BGtemplate_class;

// declare variables used by your external in the following structure
// in many of the max/msp functions below this will be handed in as
//		t_BGtemplate *x
// and you can then say
//		x->thevariable
// to get/set the values of these variables
typedef struct _BGtemplate
{
	// header
	t_pxobject x_obj;

	// variables specific to this object (see not above)
	float srate;				  	// sample rate var for the heck of it
	long num_inputs, num_outputs; 	// number of inputs and outputs
	short in_connected[MAX_INPUTS]; // we'll store whether there are 0 or more signal connections
 
	void *outpointer; // we're going to set this below to point to the right-most outlet for sending bangs and vals out
} t_BGtemplate;


/****PROTOTYPES****/

//	max/msp setup funcs; this probably won't change, unless you decide to change the number of
//	args that the user can input, in which case BGtemplate_new will have to change
void *BGtemplate_new(long num_inputs, long num_outputs);	// this allows us to have two optional params for the object
void BGtemplate_dsp(t_BGtemplate *x, t_signal **sp, short *count); 
t_int *BGtemplate_perform(t_int *w);
void BGtemplate_assist(t_BGtemplate *x, void *b, long m, long a, char *s);

// the following are functions we define below for handling floats/ints/bangs
// at the inputs.  We can delete these if we don't use them
void BGtemplate_float(t_BGtemplate *x, double f);
void BGtemplate_int(t_BGtemplate *x, int i);
void BGtemplate_bang(t_BGtemplate *x);

// function prototypes here for custom messages -- for example, the message "mymessage" will be used to call this function
// these are set up using the addmess() function in main() below
void BGtemplate_mymessage(t_BGtemplate *x, Symbol *s, short argc, Atom *argv);


/****FUNCTIONS****/

//primary MSP funcs -- this is called the first time you instantiate an object of this type
void main(void)
{
	// the two A_DEFLONG arguments give us the two optional object params, we use them to set ins/outs in BGtemplate_new() below
	// change these if you want different user args, or different types, or none... etc.
	// see the WritingExternals manual for the other max/msp datatypes (A_DEFFLOAT, A_DEFSYM, etc.)
	setup((struct messlist **)&BGtemplate_class, (method)BGtemplate_new, (method)dsp_free, (short)sizeof(t_BGtemplate), 0L, A_DEFLONG, A_DEFLONG, 0);

	// max/msp standard messages; don't change these  
	addmess((method)BGtemplate_dsp, "dsp", A_CANT, 0);
	addmess((method)BGtemplate_assist,"assist", A_CANT,0);

	// this is where we create our own messages -- the message "mysmessage" will trigger the BGtemplate_mymessage() function
	addmess((method)BGtemplate_mymessage, "mymessage", A_GIMME, 0);

	// BGtemplate_float will be called for any floats coming in any inlet
	addfloat((method)BGtemplate_float);

	// BGtemplate_int will be called for any ints coming in any inlet
	addint((method)BGtemplate_int);

	// BGtemplate_bang will be called for any bangs coming in any inlet
	addbang((method)BGtemplate_bang);

	// msp initialization (gotta have this)
	dsp_initclass();

	post("BGtemplate~ external loaded!");
}


// this gets called when the object is created and every time the user types in new args
void *BGtemplate_new(long num_inputs, long num_outputs)
{
	int i;

	// leave this; creates our object
	t_BGtemplate *x = (t_BGtemplate *)newobject(BGtemplate_class);

	// zero out the struct, to be careful (takk to jkclayton)
	if (x) { 
		for (i = sizeof(t_pxobject); i < sizeof(t_BGtemplate); i++)  
			((char *)x)[i] = 0;
	}

	// this is where we access the values (if any) of the optional params for the object	
	// constrain number of inputs and outputs
	if (num_inputs < 1) num_inputs = 1;
	if (num_inputs > MAX_INPUTS) num_inputs = 1;
	if (num_outputs < 1) num_outputs = 1;
	if (num_outputs > MAX_OUTPUTS) num_outputs = 1;

	// and we set the internal t_BGtemplate structure variables
	x->num_inputs = num_inputs;
	x->num_outputs = num_outputs;

	// now setup audio inputs and outputs
	// inputs
	dsp_setup((t_pxobject *)x, x->num_inputs);


	// here we set the 'outpointer' variable to reference the right-most outlet.  We will use this in our
	// BGtemplate_float(), BGtemplate_int(), BGtemplate_bang() functions below to send out non-signal information
	// this will be the right-most outlet because we are creating it before the signal outlets.  max/msp
	// creates oject outlets right-to-left in the order they are set up.
	x->outpointer = outlet_new((t_object *)x, 0);
   
	// signal outputs
	for (i = 0; i < x->num_outputs; i++) {
		outlet_new((t_object *)x, "signal");
	}

	//initialize input variables; important to do this!
	for (i = 0; i < x->num_inputs; i++) {
		x->in_connected[i] = 0;
	}

	//occasionally this line is necessary if you are doing weird asynchronous things with the in/out vectors
	//x->x_obj.z_misc = Z_NO_INPLACE;

	return (x);
}


// this gets called everytime audio is started; even when audio is running, if the user changes anything
// (like deletes a patch cord), audio will be turned off and then on again, calling this func.
// this adds the "perform" method to the DSP chain, and also tells us
// where the audio vectors are and how big they are
void BGtemplate_dsp(t_BGtemplate *x, t_signal **sp, short *count)
{
	void *dsp_add_args[MAX_INPUTS + MAX_OUTPUTS + 2];
	int i;

	// set sample rate var, just for fun
	x->srate = sp[0]->s_sr;

	// check to see if there are signals connected to the various inputs
	// count[] contains the number of audio signals connected to each inlet
	for (i = 0; i < x->num_inputs; i++)
		x->in_connected[i] = count[i];

	// construct the array of vectors and stuff for max/msp to knoww
	dsp_add_args[0] = x; // the object itself
	for (i = 0; i < (x->num_inputs + x->num_outputs); i++) { // pointers to the input and output vectors
		dsp_add_args[i+1] = sp[i]->s_vec;
	}
	dsp_add_args[x->num_inputs + x->num_outputs + 1] = (void *)sp[0]->s_n; // pointer to the vector size
	dsp_addv(BGtemplate_perform, (x->num_inputs + x->num_outputs + 2), dsp_add_args); // add them to the signal chain	
}


// this is where the action is
// we get vectors of samples (n samples per vector), process them and send them out
// the "w" array contains a lot of stuff:
//	w[0] -- pointer to the BGtemplate_perform function being called
//	w[1] -- pointer to the t_BGtemplate object structure data
//	w[2] ... w[numinputs+1] -- pointers to arrays of samples, one vectors'-worth for each input
//	w[numinputs+2] ... w[numinputs+numoutputs+1] -- pointers to arrays to be filled with samples (output arrays)
//	w[numinputs+numoutputs+2] -- number of samples/vector (i.e. the number of samples in each input or output array)
t_int *BGtemplate_perform(t_int *w)
{
	t_BGtemplate *x = (t_BGtemplate *)(w[1]);

	float *in[MAX_INPUTS]; 		// pointers to the input vectors
	float *out[MAX_OUTPUTS];	// pointers to the output vectors
	// note the the above are pointers -- each input/output is represented as an array
	// of samples (i.e. non-interleaved).  These are set to point to the w[] arrays.

	long n = w[x->num_inputs + x->num_outputs + 2];	// number of samples per vector

	// local vars
	int i;

	// check to see if we should skip this routine if the patcher is "muted"
	// all audio objects should have this, even though the 'goto' is supremely elegant.
	if (x->x_obj.z_disabled) goto out;

	// check to see if we have a signal connected to each input and assign the pointer accordingly
	for (i = 0; i < x->num_inputs; i++) {
		if (x->in_connected[i] > 0) in[i] = (float *)(w[i+2]);
	}

	// assign the output vector pointers
	for (i = 0 ;i < x->num_outputs; i++) {
		out[i] = (float *)(w[x->num_inputs+i+2]);
	}
	// after these two loops above, we can get samples using *in[inlet]+sampval and *out[oulet]+sampval to
	// get and set sample values for each inlet and outlet vector.


/* ------- this is the sample computing loop for each vector ------- */
	// we're going to count down by the total # of samples, and process each of the current input and current
	// output vectors.  Alternatively we could have counted through each vector separately, but we didn't
	while (n--) { // n is the number of samples for the vector

		// we'll simply transfer the inputs to the outputs, and ignore additional inputs if inlets > outlets
		// the "++" on the in[] and out[] sample vector pointers move to the next sample in the vector
		// note: for some reason signal is *required* in max/msp at the left-most inlet for processing to take place
		for (i = 0; i < x->num_inputs; i++) {
			if (x->in_connected[i] > 0) {
				if (i < x->num_outputs) {
					*out[i]++ = *in[i]++;
				}
			} else {
				if (i < x->num_outputs)
					*out[i]++ = 0.0;
			}
		}

		for (i = x->num_inputs; i < x->num_outputs; i++)
			*out[i]++ = 0.0; // if there are more outlets than inlets, we put out zeros
	}
/* ------- end of the sample computing loop ------- */


	// return a pointer to the next object in the signal chain.
out:
	return w + x->num_inputs + x->num_outputs + 3;
}


// tells the user about the inputs/outputs when mousing over them
// "m" is inlet (1) or outlet (2); "a" is which one the mouse is over
void BGtemplate_assist(t_BGtemplate *x, void *b, long m, long a, char *s)
{
	int i;
	//could use switch/case inside for loops, to give more informative assist info....
	if (m == 1) {
		for (i = 0; i< x->num_inputs; i++)
			if (a == i) sprintf(s, "input %d", i);  // this is the message that gets shown in the window frame
	}
	if (m == 2) {
		for (i = 0; i <= x->num_outputs; i++) // "<=" because we have one additional outlet for bangs/vals
			if (a == i) sprintf(s, "output %d", i);  // ditto about message that gets shown
	}
}


// this gets called whenever a float is received at *any* input
void BGtemplate_float(t_BGtemplate *x, double f)
{
	int i;

	// check to see which input the float came in -- isn't this fun?
	for (i = 0; i < x->num_inputs; i++) {
		if (x->x_obj.z_in == i) {
			// here you could set some internal variable to "f"
			post("BGtemplate~: received float number %f at inlet %d", f, i);
		}
	}

	// we'll send it out the right-most oulet (x->outpointer was set to it)
	outlet_float(x->outpointer, f);
}


// this gets called whenever an int is received at *any* input
void BGtemplate_int(t_BGtemplate *x, int k)
{
	int i;

	// check to see which input the int came in
	for (i = 0; i < x->num_inputs; i++) {
		if (x->x_obj.z_in == i) {
			// here you could set some internal variable to "k"
			post("BGtemplate~: received int number %d at inlet %d", k, i);
		}
	}

	// we'll send it out the right-most oulet (x->outpointer was set to it)
	outlet_int(x->outpointer, k);
}


// this gets called whenever a bang is received at *any* input
void BGtemplate_bang(t_BGtemplate *x)
{
	int i;

	// check to see which input the int came in
	for (i = 0; i < x->num_inputs; i++) {
		if (x->x_obj.z_in == i) {
			// here you could set some internal variable to "k"
			post("BGtemplate~: received a bang inlet %d", i);
		}
	}

	// we'll send a bang out the right-most oulet (x->outpointer was set to it)
	outlet_bang(x->outpointer);
}


//what to do when we get the message "mymessage" and a value (or list of values)
void BGtemplate_mymessage(t_BGtemplate *x, Symbol *s, short argc, Atom *argv)
{
	int i;

	// these two variables will collect either floats or ints (longs) from the message
	float temp;
	long temp2; 

	// argc is the number of arguments in the message box with "mymessage"
	for (i = 0; i < argc; i++) {

		// we'll check the type of each argument and then assign it (or in this case print it)
		switch (argv[i].a_type) {
			case A_LONG:
				temp2 = argv[i].a_w.w_long;
    			post("BGtemplate~: received argument %d of mymessage with value %d", i+1, temp2);
				break;
			case A_FLOAT:
				temp = argv[i].a_w.w_float;
    			post("BGtemplate~: received argument %d of mymessage with value %f", i+1, temp);
				break;
		}
	}
}
